import re
from html.parser import HTMLParser

FONT_SIZE_ADJUST = {
    "+1": 2,
    "-1": -2,
    "+2": 4,
    "-2": -4,
}


def parse_padding(padding_str):
    # Converts "0.05cm 0.05cm 0.05cm 0.05cm" → [top, right, bottom, left]
    sizes = [float(x.replace("cm", "")) for x in f"{padding_str}".strip().split()]
    if len(sizes) == 1:
        return [sizes[0]] * 4
    elif len(sizes) == 2:
        return [sizes[0], sizes[1], sizes[0], sizes[1]]
    elif len(sizes) == 3:
        return [sizes[0], sizes[1], sizes[2], sizes[1]]
    elif len(sizes) == 4:
        return sizes
    else:
        return [0.0, 0.0, 0.0, 0.0]


class HTMLStyledTextExtractor(HTMLParser):
    def __init__(self, base_font_size):
        super().__init__()
        self.base_font_size = base_font_size
        self.result = []  # list of (word, effective_font_size, style_factor)
        self.font_stack = [base_font_size]
        self.bold = False
        self.italic = False
        self.underline = False

    def handle_starttag(self, tag, attrs):
        if tag == "font":
            attr_dict = dict(attrs)
            size = attr_dict.get("size")
            if size in FONT_SIZE_ADJUST:
                self.font_stack.append(self.font_stack[-1] + FONT_SIZE_ADJUST[size])
            else:
                self.font_stack.append(self.font_stack[-1])
        elif tag == "b":
            self.bold = True
        elif tag == "i":
            self.italic = True
        elif tag == "u":
            self.underline = True

    def handle_endtag(self, tag):
        if tag == "font" and len(self.font_stack) > 1:
            self.font_stack.pop()
        elif tag == "b":
            self.bold = False
        elif tag == "i":
            self.italic = False
        elif tag == "u":
            self.underline = False

    def handle_data(self, data):
        words = data.split()
        for word in words:
            style_factor = 1.0
            if self.bold:
                style_factor += 0.10
            if self.italic:
                style_factor += 0.05
            # underline could affect height slightly, not width
            self.result.append((word, self.font_stack[-1], style_factor))


def estimate_cell_height_cm(cell_data):
    text = cell_data["data"]
    style = cell_data["style"]
    width_cm = float(cell_data["width"])
    padding_cm = parse_padding(style.get("padding", "0cm 0cm 0cm 0.0cm"))
    font_size_pt = int(re.search(r"\d+", str(style.get("font-size", "10pt"))).group())

    # Available text width in cm
    available_width_cm = width_cm - padding_cm[1] - padding_cm[3]
    if available_width_cm <= 0:
        return 0.0

    # Extract styled words
    parser = HTMLStyledTextExtractor(base_font_size=font_size_pt)
    parser.feed(text)
    styled_words = parser.result  # list of (word, font_size, style_factor)

    # Simulate text layout
    lines = []
    current_line_width = 0.0
    current_line = []

    for word, size_pt, style_factor in styled_words:
        avg_char_width_cm = 0.45 * size_pt * 0.0352778 * style_factor
        word_width_cm = len(word) * avg_char_width_cm

        if current_line_width + word_width_cm <= available_width_cm:
            current_line.append(word)
            current_line_width += word_width_cm + avg_char_width_cm  # add space
        else:
            lines.append(current_line)
            current_line = [word]
            current_line_width = word_width_cm + avg_char_width_cm
    if current_line:
        lines.append(current_line)

    # Estimate line height (slightly increased if underline was used)
    underline_used = any(parser.underline for (_, _, _) in styled_words)
    max_font_size = max((size for (_, size, _) in styled_words), default=font_size_pt)
    line_height_cm = max_font_size * 0.0352778 * (1.2 + (0.1 if underline_used else 0))

    total_text_height_cm = len(lines) * line_height_cm
    total_height_cm = total_text_height_cm + padding_cm[0] + padding_cm[2]

    return round(total_height_cm, 2)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from q2report.q2printer.q2printer import Q2Printer
from q2report.q2utils import int_
from q2report.q2printer.calc_height import parse_padding

import webbrowser
import os


class Q2PrinterHtml(Q2Printer):
    def __init__(self, output_file, output_type=None):
        super().__init__(output_file, output_type)
        self.html = []
        self.style = {}
        self.table_is_opened = False

    def save(self):
        self.close_html_table()
        html = (
            '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">'
            "\n<html>"
            "\n\t<head>"
            '\n\t\t<meta http-equiv="content-type" content="text/html; charset=UTF-8"/>'
            '\n\t\t<meta name="generator" content="q2report"/>'
            '\n\t\t<style type="text/css">'
        )
        html += "".join([f"\n.{self.style[x]} {x}" for x in self.style])
        html += "@media print {body {print-color-adjust: exact;}}"
        html += "\n\t\t</style>\n\t</head>\n\t<body>\n\n"

        html += "\n".join(self.html)
        html += "\n\n\t</body>\n</html>"
        if isinstance(self.output_file, str):
            with open(self.output_file, "w", encoding="utf8") as f:
                f.write(html)
        else:
            self.output_file.write(html.encode())
        return super().save()

    def reset_page(self, **args):
        super().reset_page(**args)
        size = f"{self.page_width}cm {self.page_height}cm;"
        margin = (
            f"{self.page_margin_top}cm "
            f"{self.page_margin_right}cm "
            f"{self.page_margin_bottom}cm "
            f"{self.page_margin_left}cm;"
        )
        style = '<style type="text/css"> @page {size:' + size + "margin:" + margin + "}" + "</style>"
        self.html.append(style)

    def reset_columns(self, widths=None):
        self.close_html_table()
        if widths:
            super().reset_columns(widths)
        self.open_html_table()

    def open_html_table(self):
        self.table_is_opened = True
        self.html.append('<table style="border-collapse:collapse; table-layout: fixed;">')
        self.html.append("<colgroup>")
        for col in self._cm_columns_widths:
            self.html.append(f'\t<col span="1" style="width: {col * 10}mm;"/>')
        self.html.append("</colgroup>")

    def close_html_table(self):
        if self.html and self.table_is_opened:
            self.html.append("</table>")
            self.table_is_opened = False

    def get_style_index(self, style):
        style["padding"] = " ".join([f"{x}cm" for x in parse_padding(style.get("padding", ""))])
        style_text = "; ".join([f"{x}:{style[x]}" for x in style])
        style_text = "{" + "border: solid;" + style_text + "}"
        if style_text not in self.style:
            self.style[style_text] = f"css{len(self.style)}"
        return self.style[style_text]

    def render_rows_section(self, rows_section, style, outline_level):
        super().render_rows_section(rows_section, style, outline_level)
        row_count = len(rows_section["heights"])
        spanned_cells = []
        if rows_section["role"] == "table_header":
            self.reset_columns()
            self.html.append("\t<thead>")
        for row in range(row_count):
            row_height_cm = rows_section["row_height"][row]
            # if row in rows_section["auto_height_rows"]:
            #     height = 0

            if row_height_cm != 0:
                self.html.append(f'\t<tr style="height: {row_height_cm}cm;">')
            elif rows_section["row_height"][row] == 0 and row in rows_section["hidden_rows"]:
                self.html.append('\t<tr  style="visibility:collapse">')
            else:
                self.html.append("\t<tr>")

            for col in range(self._columns_count):
                key = f"{row},{col}"
                if key in spanned_cells:
                    continue
                cell_data = rows_section.get("cells", {}).get(key, {})
                row_span = cell_data.get("rowspan", 1)
                col_span = cell_data.get("colspan", 1)
                cell_style = cell_data.get("style", {})
                if col_span > 1:
                    _cell_width = sum(self._cm_columns_widths[col : col + col_span])
                else:
                    _cell_width = self._cm_columns_widths[col]
                if row_span > 1:
                    _cell_height = sum(rows_section["row_height"][row : row + row_span])
                else:
                    _cell_height = rows_section["row_height"][row]

                cell_text = self.render_cell_images(cell_data, _cell_width, _cell_height, cell_style)
                if cell_style:
                    style_index = self.get_style_index(cell_style)
                else:
                    style_index = self.get_style_index(style)
                if row_span > 1 or col_span > 1:
                    span_text = f' colspan="{col_span}" rowspan="{row_span}"'
                    for span_row in range(int_(row_span)):
                        for span_col in range(int_(col_span)):
                            spanned_cells.append(f"{span_row + row},{span_col + col}")
                else:
                    span_text = " "

                if row_span > 1:
                    actual_row_height_cm = sum(rows_section["row_height"][row + i] for i in range(row_span))
                else:
                    actual_row_height_cm = rows_section["row_height"][row]

                self.html.append(
                    f'\t\t<td style="position: relative; overflow: visible; height:{row_height_cm}" class="{style_index}" {span_text}>{cell_text}</td>'
                )
            self.html.append("\t</tr>")
        if rows_section["role"] == "table_header":
            self.html.append("\t</thead>")
        # if rows["role"] == "table_footer":
        #     self.html.append("<thead></thead>")

    def render_cell_images(self, cell_data, cell_width, cell_height, style):
        # cell_text = cell_data.get("data", "&nbsp;")
        cell_text = cell_data.get("data", "")
        for idx, x in enumerate(cell_data.get("images", [])):
            image = x["image"]
            image_width, image_height, _ = self.prepare_image(x, cell_data.get("width"))
            image_width, image_height = float(image_width), float(image_height)
            offset_left, offset_top = self.get_image_offset(
                float(cell_width), float(cell_height), image_width, image_height, style
            )
            cell_text = f"""
                                <div style="
                                            position: absolute;
                                            top:{offset_top}cm;
                                            left:{offset_left}cm;
                                            width:{image_width}cm;
                                            height:{image_height}cm;
                                            background-image:url(data:image/jpeg;base64,{image});
                                            background-repeat: no-repeat;
                                            background-size: 100% 100%;
                                            z-index: 1;
                                ">
                                </div>{cell_text if idx == 0 else ""}"""
        return cell_text

    def show(self):
        if isinstance(self.output_file, (str, bytes, os.PathLike, int)):
            webbrowser.open_new_tab(f"file://{os.path.abspath(self.output_file)}")



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import os
from q2report.q2utils import num
import sys
import subprocess
from .calc_height import estimate_cell_height_cm
from q2report.q2printer.calc_height import parse_padding

try:
    from PyQt6.QtGui import QTextDocument
    from PyQt6.QtGui import QFont
    from PyQt6.QtWidgets import QApplication

    pyqt6_installed = True
except Exception:
    pyqt6_installed = False


cm = num(96 / num(2.54))


def get_printer(output_file, output_type=None):
    if output_type is None and isinstance(output_file, str):
        output_type = os.path.splitext(output_file)[1].lower().replace(".", "")

    if output_type == "html":
        from q2report.q2printer.q2printer_html import Q2PrinterHtml as _printer
    elif output_type == "xlsx":
        from q2report.q2printer.q2printer_xlsx import Q2PrinterXlsx as _printer
    elif output_type == "docx":
        from q2report.q2printer.q2printer_docx import Q2PrinterDocx as _printer
    elif output_type == "pdf":
        from q2report.q2printer.q2printer_pdf import Q2PrinterPdf as _printer
    else:
        raise BaseException(f"format {output_type} is not supported")
    return _printer(output_file, output_type)


class Q2Printer:
    def __init__(self, output_file, output_type=None):
        self.output_file = self.checkOutputFileName(output_file)
        if output_type is None and isinstance(self.output_file, str):
            output_type = os.path.splitext(output_file)[1]
        self.output_type = output_type.lower().replace(".", "")
        self.xmlImageList = []
        # self._OF = None
        self._columns_count = None
        self._cm_columns_widths = []
        self.q2report = None
        self._pyqt6_app = None
        if pyqt6_installed:
            if QApplication.instance() is None:
                self._pyqt6_app = QApplication([])
            else:
                self._pyqt6_app = QApplication.instance()

    def checkOutputFileName(self, fileName=""):
        if not os.path.isdir(target_folder:=os.path.dirname(fileName)):
            if target_folder:
                os.makedirs(target_folder, exist_ok=True)
        co = 0
        name, ext = os.path.splitext(fileName)
        while True:
            if os.path.isfile(fileName):
                try:
                    os.remove(fileName)
                except Exception as e:
                    co += 1
                    fileName = f"{name}{co:03d}{ext}"
                    continue
            break
        return fileName

    def open_output_file(self):
        # if not os.path.isdir(os.path.dirname(self.output_file)):
        #     os.mkdir(os.path.dirname(self.output_file))
        # self._OF = open(self.output_file, "w", encoding="utf8")
        pass

    def get_cell_height(self, cell_data):
        if pyqt6_installed:
            padding = cell_data["style"]["padding"].replace("cm", "").split(" ")
            while len(padding) < 4:
                padding += padding

            style = cell_data["style"]
            text_doc = QTextDocument()
            font_size = style.get("font-size", "0")
            if isinstance(font_size, str):
                font_size = num(font_size.replace("pt", ""))
            font = QFont(style.get("font-family", "Arial"), int(font_size))
            font.setHintingPreference(QFont.HintingPreference.PreferNoHinting)
            font.setStyleStrategy(QFont.StyleStrategy.NoAntialias | QFont.StyleStrategy.ForceOutline)
            text_doc.setDefaultFont(font)
            text_doc.setDocumentMargin(0)

            text_doc.setTextWidth(
                float(num(cell_data["width"] - num(padding[1]) - num(padding[3])) * num(cm))
            )

            style = "p {%s}" % ";".join([f"{x}:{style[x]}" for x in style])
            text_doc.setDefaultStyleSheet(style)
            text_doc.setHtml("<p>%s</p>" % cell_data["data"])
            height = round((num(text_doc.size().height()) / cm) + num(padding[0]) + num(padding[2]), 2)
            return height
        else:
            return num(estimate_cell_height_cm(cell_data))

    def get_image_offset(self, cell_width, cell_height, image_width, image_height, style):
        p_top, p_right, p_bottom, p_left = parse_padding(style.get("padding", ""))
        offset_left = 0
        offset_top = 0
        if (text_align := style.get("text-align")) == "right":
            offset_left = cell_width - image_width - p_right
        elif text_align == "center":
            offset_left = (cell_width - image_width + p_left - p_right) / 2
        elif text_align == "left":
            offset_left = p_left

        if (vert_align := style.get("vertical-align")) == "top":
            offset_top = p_top
        elif vert_align == "middle":
            offset_top = (cell_height - image_height + p_top - p_bottom) / 2
        elif vert_align == "bottom":
            offset_top = cell_height - image_height - p_bottom

        return offset_left, offset_top

    def calculate_real_sizes(self, rows_section, style):
        row_count = len(rows_section["heights"])
        spanned_cells = {}
        rows_section["row_height"] = []
        rows_section["min_row_height"] = []
        rows_section["max_row_height"] = []
        rows_section["auto_height_rows"] = []
        for row in range(row_count):
            spltd_heights = str(rows_section["heights"][row]).split("-")
            min_row_height = num(spltd_heights[0])
            max_row_height = num(spltd_heights[1]) if len(spltd_heights) == 2 else 0
            rows_section["min_row_height"].append(min_row_height)
            rows_section["max_row_height"].append(max_row_height)
            rows_section["row_height"].append(0)
            image_height = 0
            for col in range(self._columns_count):
                key = f"{row},{col}"
                cell_data = rows_section.get("cells", {}).get(key, {})

                if not cell_data:
                    continue
                if cell_data.get("colspan", 0) > 1:
                    cell_data["width"] = sum(self._cm_columns_widths[col : col + cell_data["colspan"]])
                else:
                    cell_data["width"] = self._cm_columns_widths[col]
                if cell_data.get("data"):
                    cell_data["data"] = cell_data["data"].replace("\n", "<br>")
                    cell_data["height"] = self.get_cell_height(cell_data)
                    rows_section["row_height"][row] = max(
                        cell_data["height"], rows_section["row_height"][row]
                    )
                    if key in spanned_cells:
                        spanned_cells[key] = cell_data["height"]
                    else:
                        if min_row_height == 0 and max_row_height == 0:
                            if rows_section["row_height"][row] < cell_data["height"]:
                                rows_section["row_height"][row] = cell_data["height"]

                if cell_data.get("rowspan", 0) > 1 or cell_data.get("colspan", 0) > 1:
                    spanned_cells[key] = 0

                # TODO: if background image (how to know?) - do not change height
                for image in cell_data.get("images", []):
                    w, h, i = self.prepare_image(image, cell_data.get("width"))
                    if image["height"] == 0:
                        image["height"] = max(image["height"], rows_section["row_height"][row])
                    if rows_section["row_height"][row] < h:
                        rows_section["row_height"][row] = h
                        image_height = h
            if min_row_height != 0 and rows_section["row_height"][row] < min_row_height:
                rows_section["row_height"][row] = min_row_height
            if max_row_height != 0 and rows_section["row_height"][row] > max_row_height:
                rows_section["row_height"][row] = max_row_height
            ##################
            if min_row_height != 0 and max_row_height == 0:
                rows_section["row_height"][row] = min_row_height
            elif min_row_height == 0 and max_row_height != 0:
                rows_section["row_height"][row] = min(max_row_height, rows_section["row_height"][row])
            elif min_row_height and min_row_height == max_row_height:
                rows_section["row_height"][row] = max_row_height
            ##################
            # if min_row_height == 0 and max_row_height == 0 and rows_section["row_height"][row] == 0:
            if min_row_height == 0 and max_row_height == 0 and image_height == 0:
                rows_section["auto_height_rows"].append(row)

        # calculating height for spanned cells
        rows_section["hidden_rows"] = {i for i, h in enumerate(rows_section["row_height"]) if h == 0}
        for key in spanned_cells:
            start_row = int(key.split(",")[0])
            haha = 0
            for row in range(start_row, start_row + rows_section["cells"][key]["rowspan"]):
                haha += rows_section["row_height"][row] if rows_section["row_height"][row] else num(0)
                if row in rows_section["hidden_rows"]:
                    rows_section["hidden_rows"].remove(row)
            rest = spanned_cells[key] - haha
            for uprow in range(start_row, start_row + rows_section["cells"][key]["rowspan"]):
                if rows_section["row_height"][uprow] != 0:
                    continue
                if rest > rows_section["row_height"][uprow]:
                    if rows_section["max_row_height"][uprow] == 0:
                        rows_section["row_height"][uprow] += rest
                        break
        rows_section["section_height"] = sum(rows_section["row_height"])

    def render_rows_section(self, rows_section, style, outline_level):
        self.calculate_real_sizes(rows_section, style)

    def reset_page(
        self,
        page_width=21.0,
        page_height=29.0,
        page_margin_left=2.0,
        page_margin_top=1.0,
        page_margin_right=1.0,
        page_margin_bottom=1.0,
    ):
        self.page_width = num(page_width)
        self.page_height = num(page_height)
        self.page_margin_left = num(page_margin_left)
        self.page_margin_right = num(page_margin_right)
        self.page_margin_top = num(page_margin_top)
        self.page_margin_bottom = num(page_margin_bottom)
        self._current_width = self.page_width - self.page_margin_left - self.page_margin_right
        self._current_height = self.page_height - self.page_margin_top - self.page_margin_bottom

    def reset_columns(self, widths=[]):
        self._columns_count = len(widths)
        _cm_page_width = self._current_width

        self._cm_columns_widths = [0 for x in range(self._columns_count)]

        _fixed_columns_width = [num(x) if "%" not in x and num(x) != 0 else 0 for x in widths]
        _procent_columns_width = [num(x.replace("%", "").strip()) if "%" in x else 0 for x in widths]
        _float_columns_count = (
            self._columns_count
            - len([x for x in _procent_columns_width if x != 0])
            - len([x for x in _fixed_columns_width if x != 0])
        )
        _procent_width = num((_cm_page_width - num(sum(_fixed_columns_width))) / num(100))

        for x in range(self._columns_count):
            if _fixed_columns_width[x] != 0:
                self._cm_columns_widths[x] = _fixed_columns_width[x]
            else:
                prc = _procent_columns_width[x]
                if prc == 0:
                    prc = (num(100) - sum(_procent_columns_width)) / _float_columns_count
                self._cm_columns_widths[x] = round(_procent_width * prc, 2)
        self._cm_columns_widths = [round(x, 2) for x in self._cm_columns_widths]

    def prepare_image(self, image_dict, col_width):
        # col_width = self._cm_columns_widths[col]
        image = image_dict["image"]
        width = image_dict["width"]
        height = image_dict["height"]
        pixel_width = image_dict["pixel_width"]
        pixel_height = image_dict["pixel_height"]

        if image not in self.xmlImageList:
            self.xmlImageList.append(image)
            imageIndex = len(self.xmlImageList) - 1
        else:
            imageIndex = self.xmlImageList.index(image)
        if height == 0:
            if width == 0:
                width = num(col_width)
            height = width * pixel_height / pixel_width
        elif width == 0 and height != 0:
            width = height * pixel_width / pixel_height

        return width, height, imageIndex

    def save(self):
        # self._OF.close()
        pass

    def show(self):
        if isinstance(self.output_file, (str, bytes, os.PathLike, int)):
            if os.path.isfile(self.output_file):
                if sys.platform == "win32":
                    # os.startfile(os.path.abspath(self.output_file))
                    subprocess.Popen(
                        ["start", os.path.abspath(self.output_file)],
                        close_fds=True,
                        shell=True,
                        creationflags=subprocess.DETACHED_PROCESS,
                    )
                # elif sys.platform == 'darwin':
                #     subprocess.Popen(["open", self.output_file], close_fds=True, shell=False)
                else:
                    opener = "open" if sys.platform == "darwin" else "xdg-open"
                    subprocess.Popen([opener, self.output_file], close_fds=True, shell=False)



#    Copyright (C) 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from decimal import Decimal
import datetime
import re


def is_sub_list(sublst, lst):
    return len([x for x in sublst if x in lst]) == len(sublst)


def int_(toInt):
    try:
        return int(f"{toInt}")
    except Exception:
        return int(num(toInt))


def float_(toFloat):
    try:
        return float(f"{toFloat}")
    except Exception:
        return float(num(toFloat))


def num(tonum):
    try:
        return Decimal(f"{tonum}")
    except Exception:
        return 0


def today():
    return datetime.date.today()


class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def set_dict_default(_dict, _key, _value):
    """set only key does not exist"""

    if not _dict.get(_key):
        _dict[_key] = _value


inch = num(72.0)
cm = num(28.35)
mm = cm * num(0.1)
pica = num(12.0)
pt = num(1.0)
twip = num(20.0)

reMultiSpaceDelete = re.compile(r"[\s]+")
reDecimal = re.compile(r"^[+-]?(?:\d+(?:\.\d*)?|\.\d+)$")
reNumber = re.compile(r"\d+")


class Q2Heap:
    pass
